#!/usr/bin/env Rscript
##------------------------------------------------------------------------------##
##  ATAC-seq allelic imbalance in CLL.                                          ##
##------------------------------------------------------------------------------##

#GLOBAL PARAMETERS#

#We provide the files needed for running the script as example
#Table specifying genotypes
cllGenoPath = "./data/genotypes_examples.tsv"

#Table for translating samples IDs from two different sources. 
samplesInfoPath = "./data/overview_samples_blueprint.csv"

#Directory containing BAM files processed for dealing with mapping bias with 
#WASP. These data must be obtained using the WASP pipeline. 
samplesDir = "/gpfs/scratch/hcli48/hcli48037/results/manuscripts/Risk_SNPs_ICRL/allelic_imbalance/filtered_bams"

#File containing SNPs for overdispersion estimation
bkdSnpsPath = "./data/atac_bkd_snps.tsv"

#File containing target SNPs list 
targetSnpsPath = "./data/225_SNPs_allelic_imbalance.tsv"

#Imprinted genes
imprGenesPath = "./data/imprinted_genes.tsv"

#Output directory for results
outDir = "./results"

#GLOBAL PARAMETERS#

suppressPackageStartupMessages(library(GenomicRanges))
suppressPackageStartupMessages(library(Rsamtools))
suppressPackageStartupMessages(library(parallel))

#Functions
#Pools samples counts at snps
snpsAcc <- function(mark) {
  
  cat(sprintf("Processing %s\n", mark[1]))
  
  #ATAC-seq
  counts = matrix(0, nrow = nrow(snps), ncol = 5)
  rownames(counts) = snps[,2]
  colnames(counts) = c('A', 'C', 'G', 'T', 'other')
  
  counts3d = array(0, dim = c(nrow(snps), 5, nrow(cllGeno)),
                   dimnames = list(snps[,2], c('A', 'C', 'G', 'T', 'other'),
                                   rownames(cllGeno)))
  
  cSamplesSnp = rep(0, nrow(snps))
  names(cSamplesSnp) = snps[,2]
  
  cSamplesWithReadsSnp = rep(0, nrow(snps))
  names(cSamplesWithReadsSnp) = snps[,2]
  
  cntsSmp = mclapply(1:nrow(cllGeno), function(x) 
    pileupSample(x, cllGeno, samplesInfo, snps), 
                     mc.cores = 7)
  #Summarize results
  idsCll = rownames(cllGeno)
  for (i in 1:nrow(cllGeno)) {
    counts = counts + cntsSmp[[i]]$counts
    counts3d[,,idsCll[i]] = cntsSmp[[i]]$counts
    cSamplesSnp = cSamplesSnp + cntsSmp[[i]]$cSamplesSnp
    cSamplesWithReadsSnp = cSamplesWithReadsSnp + cntsSmp[[i]]$cSamplesWithReadsSnp
  }

  return (list(counts = counts, counts3d = counts3d, 
               cSamplesSnp = cSamplesSnp, 
               cSamplesWithReadsSnp = cSamplesWithReadsSnp))
}

#Gets pile-up information from one sample at het SNPs
pileupSample <- function(iCll, cllGeno, samplesInfo, snps) {
  
  st = Sys.time()
  
  counts = matrix(0, nrow = nrow(snps), ncol = 5)
  rownames(counts) = snps[,2]
  colnames(counts) = c('A', 'C', 'G', 'T', 'other')
  
  cSamplesSnp = rep(0, nrow(snps))
  names(cSamplesSnp) = snps[,2]
  
  cSamplesWithReadsSnp = rep(0, nrow(snps))
  names(cSamplesWithReadsSnp) = snps[,2]
  
  jHet = which(cllGeno[iCll,] == "het")
  
  #Retrieve file
  id1Cll = rownames(cllGeno)[iCll]
  id2Cll = samplesInfo[id1Cll, mark[1]]
  
  cat(sprintf("Processing %s - %s\n", id1Cll, id2Cll))
  
  samplePath = sprintf("%s/%s", samplesDir,
                       list.files(samplesDir, sprintf(".*%s.*%s.*.bam$", 
                                                      id2Cll, mark[3])))
  samplePath = samplePath[grep("filtered.sorted", samplePath, invert = T)]
  
  if (length(samplePath) == 1) {
    
    iSnp = match(colnames(cllGeno)[jHet], snps[,2])
    iSnp = iSnp[!is.na(iSnp)]
    
    snpsLoc = data.frame(chr = snps[iSnp, 3], start = snps[iSnp, 4] + 1,
                         end = snps[iSnp, 4] + 1, dbSNP = snps[iSnp, 2],
                         stringsAsFactors = F)
    if (nrow(snpsLoc) > 0) {
      rownames(snpsLoc) = sprintf("%s:%d-%d", snpsLoc$chr, snpsLoc$start, snpsLoc$end)
      snpsLocGr = GRanges(snpsLoc)
      
      #Extract pile-up & nucleotide stats
      bf = BamFile(samplePath)
      param = ScanBamParam(which=snpsLocGr) 
      paramPileup =  PileupParam(max_depth = 1000, min_base_quality = 5, 
                                 distinguish_strands = F)
      snpsPileup = pileup(bf, scanBamParam=param, pileupParam = paramPileup)
      
      lastSnpId = ''
      for (iRes in 1:nrow(snpsPileup)) {
        snpId = snpsLoc[as.character(snpsPileup[iRes, "which_label"]), "dbSNP"]
        
        if (snpId != lastSnpId || iRes == nrow(snpsPileup)) {
          if  (lastSnpId != '') {
            
            if (iRes == nrow(snpsPileup)) {
              nucl = toupper(as.character(snpsPileup[iRes, "nucleotide"]))
              nucl = ifelse(nucl == "A" || nucl == "C" || nucl == "G" || nucl == "T", 
                            nucl, "other")
              counts[snpId, nucl] = counts[snpId, nucl] + 
                as.numeric(snpsPileup[iRes, "count"])
              
              cNuclReads[nucl] = cNuclReads[nucl] + as.numeric(snpsPileup[iRes, "count"])
            }
            
            snpsAlleles = unlist(strsplit(sub('\\)', '', 
                                              sub('\\(', '', snps[lastSnpId,5])),
                                          '/'))
            allelesCount = cNuclReads[snpsAlleles]
            
            if (sum(allelesCount)) {
              cSamplesWithReadsSnp[lastSnpId] = cSamplesWithReadsSnp[lastSnpId] + 1
            }
          }
          
          lastSnpId = snpId
          cNuclReads = rep(0, 5)
          names(cNuclReads) = c('A','C','G','T','other')
        }
        
        nucl = toupper(as.character(snpsPileup[iRes, "nucleotide"]))
        nucl = ifelse(nucl == "A" || nucl == "C" || nucl == "G" || nucl == "T", 
                      nucl, "other")
        
        counts[snpId, nucl] = counts[snpId, nucl] + 
          as.numeric(snpsPileup[iRes, "count"])
        cNuclReads[nucl] = cNuclReads[nucl] + 
          as.numeric(snpsPileup[iRes, "count"])
      }
      cSamplesSnp[colnames(cllGeno)[jHet]] = cSamplesSnp[colnames(cllGeno)[jHet]] + 1
    }
  } else {
    cat(sprintf("Warning! Could not properly retrieve sample file for %s-%s %s: %d files found\n",
                id1Cll, id2Cll, mark[1], length(samplePath)))
  }
  
  cat(sprintf("%s-%s: ", id1Cll, id2Cll), Sys.time() - st, "\n",
      file = stderr())
  
  return (list(counts = counts, cSamplesSnp = cSamplesSnp, 
               cSamplesWithReadsSnp = cSamplesWithReadsSnp))
}

#Two-tailed test from Beta-Binomial 
pvalBetaBinomial2tail <- function(allelesCount, fit) {
  
  n = sum(allelesCount)
  k = min(allelesCount)
  
  #Probability lower equal
  pval = min(1, 2*pbetabinom.ab(k, n, fit["shape1"], fit["shape2"]))
  
  return(pval)
}

#Main
cat("Initializing\n")
st = Sys.time()
mark = c("ATAC-seq", "atac", "ATAC")

cllGeno = read.table(cllGenoPath, stringsAsFactors = F)
samplesInfo = read.delim(samplesInfoPath, stringsAsFactors = F, check.names = F)
rownames(samplesInfo) = samplesInfo$sample

print(Sys.time() - st)
  
snps = read.delim(bkdSnpsPath, header = F, stringsAsFactors = F)
rownames(snps) = snps[,2]

resPileup = snpsAcc(mark)

cat("Beta-binomial\n")

cSnps = nrow(resPileup$counts)
resAllImb = data.frame(mark = rep('', cSnps), snp = rep('', cSnps),
                       chr = rep('', cSnps), pos = rep(NA, cSnps), A = rep(NA, cSnps),
                       C = rep(NA, cSnps), G = rep(NA, cSnps), `T` = rep(NA, cSnps),
                       heterozygous_samples = rep(NA, cSnps),
                       with_reads = rep(NA, cSnps),
                       excluded = rep(NA, cSnps),
                       included = rep(NA, cSnps),
                       A_incl = rep(NA, cSnps), C_incl = rep(NA, cSnps),
                       G_incl = rep(NA, cSnps), T_incl = rep(NA, cSnps),
                       fImbalance_bb = rep('', cSnps), pval_bb = rep(NA, cSnps),
                       fdr_bb = rep(NA, cSnps), abundant_bb = rep('', cSnps),
                       ref_count = rep(NA, cSnps), alt_count = rep(NA, cSnps),
                       stringsAsFactors = F)
st = Sys.time()

for (iSnp in 1:nrow(resPileup$counts)) {
  
  # st = Sys.time()

  idSnp = rownames(resPileup$counts)[iSnp]

  counts = resPileup$counts[iSnp, c('A','C','G','T')]
  resSnp = c(mark[1], idSnp, snps[idSnp, 3], snps[idSnp, 4] + 1, counts,
             0, 0, 0, 0, 0, 0, 0, 0, NA, NA, NA, NA, 0, 0)

  resSnp[9] = resPileup$cSamplesSnp[idSnp]
  resSnp[10] = resPileup$cSamplesWithReadsSnp[idSnp]

  alleles = snps[idSnp, 5]
  alleleA = toupper(sub('\\(', '', strsplit(alleles, '/')[[1]][1]))
  alleleB = toupper(sub('\\)', '', strsplit(alleles, '/')[[1]][2]))

  countsSamples = resPileup$counts3d[iSnp, c('A','C','G','T'), ]
  jHetCllWithReads = which(colSums(countsSamples[c(alleleA, alleleB),]) > 0)
  
  #Only samples which observed alleles ratio is between 0.1 & 0.9
  cA = countsSamples[alleleA, jHetCllWithReads]
  cB = countsSamples[alleleB, jHetCllWithReads]
  ratio = cA / (cA + cB)
  jExcl = which(ratio <= 0.1 | ratio >=  0.9)
  jIncl = jHetCllWithReads[setdiff(1:length(jHetCllWithReads), jExcl)]

  resSnp[11] = length(jExcl)
  resSnp[12] = as.numeric(resSnp[10]) - as.numeric(resSnp[11])
  resSnp[13] = sum(countsSamples['A', jIncl])
  resSnp[14] = sum(countsSamples['C', jIncl])
  resSnp[15] = sum(countsSamples['G', jIncl])
  resSnp[16] = sum(countsSamples['T', jIncl])

  resSnp[21] = sum(countsSamples[alleleA, jIncl])
  resSnp[22] = sum(countsSamples[alleleB, jIncl])

  resAllImb[iSnp,] = resSnp
  if (iSnp %% 5000 == 0) {
    cat(iSnp, " snps processed in ", Sys.time() - st, "\n")
  }
} 

#Beta-binomial
rownames(resAllImb) = resAllImb$snp

#Filtering target SNPs 
targetSnps = read.delim(targetSnpsPath, stringsAsFactors = F ,header = F)[,1]
pttnTgt = sprintf("^(%s)$", paste(targetSnps, collapse = "|"))
iTgtRes = grep(pttnTgt, resAllImb$snp, perl = T)
resAllImbTgtSnps = resAllImb[iTgtRes,]
rownames(resAllImbTgtSnps) = resAllImbTgtSnps$snp

#IGH, IGK, IGL and HLA loci (Human Genome Region MHC: chr6:28,510,120-33,480,577)
igGr = GRanges(data.frame(chr = c("chr14", "chr2", "chr22", "chr6"), 
                          start = c(105550000, 88816899, 22020000, 28510120), 
                          end = c(107043718, 90235370, 22962010, 33480577)))
bkdGr = GRanges(data.frame(chr=resAllImb$chr,
                           start=resAllImb$pos,
                           end=resAllImb$pos))
iBkdIg = unique(queryHits(findOverlaps(bkdGr, igGr)))

#Filtering SNPs overlapping imprinted genes
imprinted = read.delim(imprGenesPath, header = T, stringsAsFactors = F)
iChrOk = grep("^\\d\\d?$", imprinted$Chromosome.scaffold.name, perl = T)
imprinted = imprinted[iChrOk,]
imprinted$Chromosome.scaffold.name = sprintf("chr%d",
                                             as.numeric(imprinted$Chromosome.scaffold.name))
imprDf = data.frame(chr = imprinted$Chromosome.scaffold.name,
                    start = imprinted$Gene.start..bp.,
                    end = imprinted$Gene.end..bp.)
suppressPackageStartupMessages(library(GenomicRanges))
imprGr = GRanges(imprDf)
offset = 1e5
start(imprGr) = pmax(1, start(imprGr) - offset)
end(imprGr) = end(imprGr) + offset

iBkdImpr = unique(queryHits(findOverlaps(bkdGr, imprGr)))

#Filtering
iFilt = union(iBkdIg, union(iTgtRes, iBkdImpr))
resAllImb = resAllImb[-iFilt,]

cat("Beta-binomial overdispersion estimation\n")
st = Sys.time()

#Plot of all snps ratio with reads
cRef = as.numeric(resAllImb$ref_count)
cAlt = as.numeric(resAllImb$alt_count)
ratio = cRef / (cRef+cAlt)

#Snps with reads in at least 3 heterozygous samples and 10 reads 
minCountsAlleles = 10
cRef = as.numeric(resAllImb$ref_count)
cAlt = as.numeric(resAllImb$alt_count)
ratio = cRef / (cRef+cAlt)
idSnpsWithReads = resAllImb$snp[
  which(as.numeric(resAllImb$included) > 2 & 
          (cRef + cAlt) >= minCountsAlleles)]

names(ratio) = resAllImb$snp

cRefTgt = as.numeric(resAllImbTgtSnps$ref_count)
cAltTgt = as.numeric(resAllImbTgtSnps$alt_count)
ratioTgt = cRefTgt / (cRefTgt+cAltTgt)
idTgtSnpsWithReads = resAllImbTgtSnps$snp[
  which(as.numeric(resAllImbTgtSnps$included) > 2 &
          (cRefTgt + cAltTgt) >= minCountsAlleles)]

rownames(resAllImb) = resAllImb$snp

m = as.numeric(resAllImb[idSnpsWithReads,"ref_count"])
n = as.numeric(resAllImb[idSnpsWithReads,"alt_count"]) +
  as.numeric(resAllImb[idSnpsWithReads,"ref_count"])
df = data.frame(n = n, m = m)

#VGAM
library(VGAM)
mleff = VGAM::vglm(cbind(m, n-m) ~ 1, betabinomialff, 
           data = df, trace = T)
fitff = Coef(mleff)
cat(sprintf("Fit: alpha = %.3f, beta = %.4f\n", fitff[["shape1"]], fitff[["shape2"]]))

for (idSnp in idTgtSnpsWithReads) {
  
  countsSamples = as.numeric(c(resAllImbTgtSnps[idSnp, "ref_count"], 
                               resAllImbTgtSnps[idSnp, "alt_count"]))
  #Binomial test
  resAllImbTgtSnps[idSnp, "pval_bb"] =
    pvalBetaBinomial2tail(countsSamples, fitff)
}

resAllImbTgtSnps$fdr_bb = p.adjust(resAllImbTgtSnps$pval_bb, "BH")

iTests = which(!is.na(resAllImbTgtSnps$pval_bb))
iImb = which(as.double(resAllImbTgtSnps[iTests, "fdr_bb"]) < 0.1)
resAllImbTgtSnps[iTests[iImb], "fImbalance_bb"] = "YES"
resAllImbTgtSnps[iTests[-iImb], "fImbalance_bb"] = "NO"
resAllImbTgtSnps[-iTests, "fImbalance_bb"] = "NA"
resAllImbTgtSnps[iTests[iImb], 'abundant_bb'] =
  c('A','C','G','T')[apply(resAllImbTgtSnps[iTests[iImb], c('A','C','G','T')], 1, which.max)]

n = as.numeric(resAllImbTgtSnps$ref_count[iTests]) + 
  as.numeric(resAllImbTgtSnps$alt_count[iTests])
p = as.numeric(resAllImbTgtSnps$ref_count[iTests]) / n
gamm = 1 / (1 +  fitff[["shape1"]] + fitff[["shape2"]])
i0 = pmax(0, pmin(1, p + -1 * qnorm(0.975) * sqrt((1/n) * p * (1-p) * (1+(n-1)*gamm))))
i1 = pmax(0, pmin(1, p +  1 * qnorm(0.975) * sqrt((1/n) * p * (1-p) * (1+(n-1)*gamm))))

resAllImbTgtSnps[, "ref_ratio"] = NA
resAllImbTgtSnps[iTests, "ref_ratio"] = p
resAllImbTgtSnps[, "Wald_CI_95"] = NA
resAllImbTgtSnps[iTests, "Wald_CI_95"] = sprintf("[%.3f, %.3f]", i0, i1)

dstPath = sprintf("%s/allelic_imbalance.tsv", outDir)
write.table(resAllImbTgtSnps, dstPath, quote = F, sep = '\t', row.names = F, col.names = T)

cat("done\n")
